/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WFUN0ELE_H
#define WFUN0ELE_H

#include "flow.h"
#include "wFct0.h"
#include "wElement.h"
#include <vector>

namespace flow
{

/**
 * @brief Nonlinear density
 * 
 * The density is limited (Padé approximation) for large value of velocity and
 * particularized for diatomic gas
 * @authors Adrien Crovato
 */
class FLOW_API F0ElRho : public tbox::Fct0
{
    double gamma; ///< heat capacity ratio (diatomic gas only)
    double mInf;  ///< freestream Mach number
    double mC2;   ///< critical Mach number

public:
    F0ElRho(double _mInf, double _mC2) : Fct0(), gamma(1.4), mInf(_mInf), mC2(_mC2) {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Linear density (constant)
 * @authors Adrien Crovato
 */
class FLOW_API F0ElRhoL : public tbox::Fct0
{
public:
    F0ElRhoL() : Fct0() {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Nonlinear Mach number
 * 
 * Particularized for diatomic gas
 * @authors Adrien Crovato
 */
class FLOW_API F0ElMach : public tbox::Fct0
{
    double gamma; ///< heat capacity ratio (diatomic gas only)
    double mInf;  ///< freestream Mach number

public:
    F0ElMach(double _mInf) : Fct0(), gamma(1.4), mInf(_mInf) {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Linear Mach number (constant)
 * @authors Adrien Crovato
 */
class FLOW_API F0ElMachL : public tbox::Fct0
{
public:
    F0ElMachL() : Fct0() {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Nonlinear pressure coefficient
 * 
 * Particularized for diatomic gas
 * @authors Adrien Crovato
 */
class FLOW_API F0ElCp : public tbox::Fct0
{
    double gamma; ///< heat capacity ratio (diatomic gas only)
    double mInf;  ///< freestream Mach number

public:
    F0ElCp(double _mInf) : Fct0(), gamma(1.4), mInf(_mInf) {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Linear pressure coefficient
 * @authors Adrien Crovato
 */
class FLOW_API F0ElCpL : public tbox::Fct0
{

public:
    F0ElCpL() : Fct0() {}

#ifndef SWIG
    virtual double eval(tbox::Element &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalD(tbox::Element &e,
                         std::vector<double> const &u,
                         size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace flow

#endif //WFUN0ELE_H
