#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Compute lifting (linear or nonlinear) viscous flow around a naca0012
# @authors Adrien Crovato, Amaury Bilocq
# Test the viscous-inviscid interaction scheme
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import flow as flo
import flow.utils as floU
import flow.default as floD
import flow.viscous.solver as floVS
import flow.viscous.coupler as floVC
import tbox
import tbox.utils as tboxU
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 0*math.pi/180
    U_inf = [math.cos(alpha), math.sin(alpha)] # norm should be 1
    M_inf = 0.0
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    c_ref = 1
    dim = len(U_inf)

    # mesh the geometry
    print ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.0075, 'msLe' : 0.0075}
    msh = gmsh.MeshLoader("../models/n0012.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack('wake')
    mshCrck.addBoundaries(['field', 'airfoil', 'downstream'])
    mshCrck.run()
    del mshCrck
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl = flo.Problem(msh, dim, alpha, M_inf, c_ref, c_ref, 0.25, 0., 0.)
    if M_inf == 0:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRhoL(), flo.F0ElMachL(), flo.F0ElCpL(), flo.F0PsPhiInf(dim, alpha)))
    else:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Initial(msh, 'field', flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Dirichlet(msh, 'upstream', flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Freestream(msh, 'side', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl.add(flo.Freestream(msh, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], 0.)))
    pbl.add(flo.Wake(msh, ['wake', 'wake_', 'field']))
    pbl.add(flo.Kutta(msh, ['te', 'wake_', 'airfoil', 'field']))
    bnd = flo.Boundary(msh, ['airfoil', 'field'])
    pbl.add(bnd)
    # TODO also add Blowing on wake...
    blw = flo.Blowing(msh, ['airfoil', 'field'])
    pbl.add(blw)
    tms['pre'].stop()

    # solve inviscid problem
    print ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET
    tms['solver'].start()
    isolver = flo.Newton(pbl)
    floD.initNewton(isolver)
    vsolver = floVS.Solver(blw)
    coupler = floVC.Coupler(isolver, vsolver, gmshWriter)
    coupler.run()
    tms['solver'].stop()

    # extract Cp
    Cp = floU.extract(bnd.groups[0].tag.elems, isolver.Cp)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    # display results
    print ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET
    print '       M    alpha       Cl       Cd       Cm'
    print '{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alpha*180/math.pi, isolver.Cl, isolver.Cd, isolver.Cm)

    # display timers
    tms['total'].stop()
    print ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET
    print 'CPU statistics'
    print tms

    # visualize solution and plot results
    floD.initViewer(pbl)
    tboxU.plot(Cp[:,0], Cp[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(isolver.Cl, isolver.Cd, isolver.Cm, 4), True)

    # check results
    print ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET
    tests = CTests()
    if M_inf == 0 and alpha == 0*math.pi/180:
        tests.add(CTest('min(Cp)', min(Cp[:,3]), -0.41, 1e-1)) # TODO check value and tolerance
        tests.add(CTest('Cl', isolver.Cl, 0., 5e-2))
    else:
        raise Exception('Test not defined for this flow')
    tests.run()

    # eof
    print ''

if __name__ == "__main__":
    main()
