#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Compute the (non)linear flow around 3D cylinder, with various B.C.
# @authors Adrien Crovato
# Test solver convergence and boundary conditions
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import flow as flo
import flow.utils as floU
import flow.default as floD
import tbox
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3*math.pi/180
    U_inf = [math.cos(alpha), 0., math.sin(alpha)] # norm should be 1
    M_inf = 0.1
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    dim = len(U_inf)

    # define dimension and mesh size
    dmt = 1.0 # cylinder diameter
    spn = 2.0 # cylinder span
    tpt = 5e-2 # cylinder tip thickness (ratio wrt span)
    lgt = 8.0 # channel length
    hgt = 6.0 # channel height
    wdt = 3.0 # channel width
    l_ref = dmt # reference length
    S_ref = dmt*spn # reference area
    fms = 0.5 # farfield mesh size
    nms = 0.1 # nearfield mesh size  

    # mesh the geometry
    print ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET
    tms['msh'].start()
    pars = {'dmt' : dmt, 'spn' : spn, 'tpt' : tpt, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msN' : nms, 'msF' : fms}
    msh = gmsh.MeshLoader("../models/cylinder_3.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack('wake')
    mshCrck.addBoundaries(['field', 'cylinder', 'symmetry', 'downstream'])
    mshCrck.setExcluded('wakeTip')
    mshCrck.run()
    del mshCrck
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl = flo.Problem(msh, dim, alpha, M_inf, S_ref, l_ref, 0., 0., 0.)
    if M_inf == 0:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRhoL(), flo.F0ElMachL(), flo.F0ElCpL(), flo.F0PsPhiInf(dim, alpha)))
    else:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha)))
    pbl.add(flo.Initial(msh, 'field', flo.F0PsPhiInf(dim, alpha)))
    dirichlet = flo.Dirichlet(msh, 'upstream', flo.F0PsPhiInf(dim, alpha))
    pbl.add(dirichlet)
    pbl.add(flo.Freestream(msh, 'farfield', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    pbl.add(flo.Freestream(msh, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    wake = flo.Wake(msh, ['wake', 'wake_', 'field', 'teTip'])
    pbl.add(wake)
    bnd = flo.Boundary(msh, ['cylinder', 'field'])
    pbl.add(bnd)
    tms['pre'].stop()

    # solve problem
    print ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET
    tms['solver'].start()
    solver = flo.Newton(pbl)
    floD.initNewton(solver)
    cnvrgd = solver.run()
    solver.save(0, gmshWriter)
    tms['solver'].stop()

    # compute mean error in Boundary, Neumann boundary condition (works only for Tri3 elements)
    gradn = []
    for e in bnd.groups[0].tag.elems:
        gradx = 0
        grady = 0
        gradz = 0
        ux = e.nodes[1].pos.x[0] - e.nodes[0].pos.x[0]
        uy = e.nodes[1].pos.x[1] - e.nodes[0].pos.x[1]
        uz = e.nodes[1].pos.x[2] - e.nodes[0].pos.x[2]
        vx = e.nodes[2].pos.x[0] - e.nodes[0].pos.x[0]
        vy = e.nodes[2].pos.x[1] - e.nodes[0].pos.x[1]
        vz = e.nodes[2].pos.x[2] - e.nodes[0].pos.x[2]
        nx = uy*vz - uz*vy
        ny = uz*vx - ux*vz
        nz = ux*vy - uy*vx
        for n in e.nodes:
            gradx += solver.U[n.row].x[0] / e.nodes.size()
            grady += solver.U[n.row].x[1] / e.nodes.size()
            gradz += solver.U[n.row].x[2] / e.nodes.size()
        gradn_ = gradx * nx + grady * ny + gradz * nz
        gradn.append(gradn_)
    errNeu = sum(gradn) / len(gradn)
    # compute largest error in Dirichlet boundary condition
    errDir = 0.
    for n in dirichlet.nodes:
        errCur = solver.phi[n.row] - (n.pos.x[0]*U_inf[0] + n.pos.x[1]*U_inf[1] + n.pos.x[2]*U_inf[2])
        if (abs(errCur) > errDir):
            errDir = abs(errCur)
    # compute largest error in Wake boundary condition
    errWak = 0.
    for n in wake.nodMap:
        errCur = solver.Cp[n.row] - solver.Cp[wake.nodMap[n].row]
        if (abs(errCur) > errWak):
            errWak = abs(errCur)

    # display timers
    tms['total'].stop()
    print ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET
    print 'CPU statistics'
    print tms

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET
    if (not cnvrgd):
        raise Exception(ccolors.ANSI_RED + 'Flow solver failed to converge!' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('Neumann B.C. mean error', errNeu, 0., 1e-2))
    tests.add(CTest('Dirichlet B.C. max error', errDir, 0., 1e-12))
    tests.add(CTest('Wake/Kutta B.C. max error', errWak, 0., 1e-1))
    tests.run()

    # eof
    print ''

if __name__ == "__main__":
    main()
