#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Compute the flow on a deformed mesh
# @authors Adrien Crovato
# Test the mesh deformation process
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import numpy as np
import flow as flo
import flow.utils as floU
import flow.default as floD
import tbox
import tbox.gmsh as gmsh
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha0 = 0 # must be zero for this testfile
    U_inf = [np.cos(alpha0), 0., np.sin(alpha0)] # norm should be 1
    M_inf = 0.3
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    dim = len(U_inf)

    # define dimension and mesh size
    spn = 1.0 # wing span
    lgt = 3.0 # channel half length
    hgt = 3.0 # channel half height
    wdt = 3.0 # channel width
    S_ref = 1.*spn # reference area
    c_ref = 1 # reference chord
    fms = 1.0 # farfield mesh size
    nms = 0.05 # nearfield mesh size 

    # parameters for mesh deformation
    alfa = 3*np.pi/180
    xc = 0.25
    dz_max = 0.2*spn

    # mesh an airfoil
    print ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET
    tms['msh'].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msLe' : nms, 'msTe' : nms, 'msF' : fms}
    msh = gmsh.MeshLoader("../models/n0012_3.geo",__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack('wake')
    mshCrck.addBoundaries(['field', 'wing', 'symmetry', 'downstream'])
    mshCrck.setExcluded('wakeTip')
    mshCrck.run()
    del mshCrck
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    tms['msh'].stop()
    # create mesh deformation handler
    mshDef = tbox.MshDeform(msh, dim)
    mshDef.setField("field")
    mshDef.addFixed(["upstream", "farfield", "downstream"])
    mshDef.addMoving(["wing"])
    mshDef.setSymmetry("symmetry", 1)
    mshDef.addInternal(["wake", "wake_"])
    mshDef.initialize()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl = flo.Problem(msh, dim, alpha0, M_inf, S_ref, c_ref, 0., 0., 0.)
    if M_inf == 0:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRhoL(), flo.F0ElMachL(), flo.F0ElCpL(), flo.F0PsPhiInf(dim, alpha0)))
    else:
        pbl.set(flo.Medium(msh, 'field', flo.F0ElRho(M_inf, M_crit), flo.F0ElMach(M_inf), flo.F0ElCp(M_inf), flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Initial(msh, 'field', flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Dirichlet(msh, 'upstream', flo.F0PsPhiInf(dim, alpha0)))
    pbl.add(flo.Freestream(msh, 'farfield', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    pbl.add(flo.Freestream(msh, 'downstream', tbox.Fct1C(-U_inf[0], -U_inf[1], -U_inf[2])))
    pbl.add(flo.Wake(msh, ['wake', 'wake_', 'field', 'teTip']))
    bnd = flo.Boundary(msh, ['wing', 'field'])
    pbl.add(bnd)
    tms['pre'].stop()

    # solver problem for 0°AOA
    print ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET
    tms['solver0'].start()
    solver = flo.Newton(pbl)
    floD.initNewton(solver)
    solver.run()
    tms['solver0'].stop()

    # deform the mesh (dummy rotation + bending)
    print ccolors.ANSI_BLUE + 'PyDeforming...' + ccolors.ANSI_RESET
    tms['deform'].start()
    floD.initMeshDeformer(mshDef)
    mshDef.savePos()
    rMat = np.matrix([[np.cos(alfa), np.sin(alfa)], [-np.sin(alfa), np.cos(alfa)]])
    vNods = {}
    for e in bnd.groups[0].tag.elems:
        for n in e.nodes:
            if n.no not in vNods:
                vNods[n.no] = n
    for no in vNods:
        nod = np.matrix([[vNods[no].pos.x[0]-xc], [vNods[no].pos.x[2]]])
        new = rMat * nod
        vNods[no].pos.x[0] = new[0,0]
        vNods[no].pos.x[2] = new[1,0] + dz_max/(spn*spn)*(vNods[no].pos.x[1]*vNods[no].pos.x[1])
    # deform the mesh
    mshDef.deform()
    gmshWriter.save(msh.name+"_def")
    tms['deform'].stop()

    # solve problem for alfa AOA
    print ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET
    tms['solver1'].start()
    solver.run()
    solver.save(0, gmshWriter)
    tms['solver1'].stop()

    # display results
    print ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET
    print '  case        M    alpha       Cl       Cd       Cm'
    print '  true {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver.Cl, solver.Cd, solver.Cm)

    # display timers
    tms['total'].stop()
    print ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET
    print 'CPU statistics'
    print tms

    # check results
    print ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET
    tests = CTests()
    if alfa == 3*np.pi/180 and M_inf == 0.3 and spn == 1.0:
        tests.add(CTest('CL', solver.Cl, 0.135, 5e-1))
        tests.add(CTest('CD', solver.Cd, 0.0062, 1e-2)) # Tranair (NF=0.0062, FF=0.0030), Panair 0.0035
    tests.run()

    # eof
    print ''

if __name__ == "__main__":
    main()
