#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# test of Neumann BCs

import heat as h
import heat.utils as u
import tbox
import tbox.gmsh as gmsh
from fwk.testing import *

class MyFlux(tbox.Fct0XYZ):
    def __init__(self, ampl, cx, cy, freq):
        tbox.Fct0XYZ.__init__(self)
        self.freq = freq
        self.ampl = ampl
        self.cx   = cx
        self.cy   = cy
    def eval(self, pt):
        import math
        x = pt.x[0]-self.cx
        y = pt.x[1]-self.cy
        return self.ampl*math.sin(math.pi*self.freq*math.sqrt(x*x+y*y)/1.)


def main():
    # mesh a unit square
    pars={ 'Lx' : 1 , 'Ly' : 1, 'Nx' : 50, 'Ny' : 50}
    msh = gmsh.MeshLoader("../../models/rect_stru.geo",__file__).execute(**pars)
    gmshWriter = tbox.GmshExport(msh)
    pbl = h.Problem(msh)

    # medium/boundary conditions
    pbl.add(h.Medium(msh, "S1", tbox.Fct2C(1.0,1.0,0.0)))
    f1 = MyFlux(100, 0, 0, 2.0)
    f2 = MyFlux(100, 1, 1, 2.0)
    pbl.add(h.Source(msh, "Left", f1))    # <== Neumann
    pbl.add(h.Source(msh, "Right", f2))   # <== Neumann
    pbl.add(h.Dirichlet(msh, "Bottom", 0.))
    pbl.add(h.Dirichlet(msh, "Top", 0.))

    # solve problem
    solver = u.solve(pbl, gmshWriter)

    # extract some results
    x1,y1,T1 = u.getTemp(solver,"Left", sort='y')
    
    u.simpleplot(y1, T1, xlabel='y [mm]', title='Temperature along y')
       
    tests = CTests()
    tests.add(CTest('mean(Tleft)', sum(T1)/len(T1), 0.0, 1e-6))
    tests.add(CTest('max(Tleft)', max(T1), 15.782894, 1e-3))
    tests.add(CTest('max+min(Tleft)', max(T1)+min(T1), 0.0, 1e-6))
    tests.run()


if __name__ == "__main__":
    main()
