#!/usr/bin/env python
# -*- coding: latin-1; -*-

import mrstlnos as m
import tbox
import tbox.gmsh as gmsh

def analytical(E,nu,L,W,H,d,p,x):
    
    EI =  (E * W * H**3)/12.
    q = p * W
    q_1 = 24*EI*d/L**4
    q_2 = 72*EI*d/L**4
    a = (72*EI*d/q)**(0.25)    

    if q < q_1:
        case_id = 1
        sol = -(q/(24*EI))*(x**4-4.*L*x**3+4.*L**2*x**2)
    if q >= q_1:
        case_id = 2
        sol = -(q/(24*EI))*(x**4-(8./3.)*a*x**3+2.*a**2*x**2)
        for i in range(0,len(x)):
            if x[i] >= a:
                sol[i] = -d            
 
    return case_id, sol

def analytical_ensemble(E,nu,L,W,H,d,p,x,ensemble_size):
    
    import numpy as np
    
    case_id = np.empty([ensemble_size,1])
    sol = np.empty([len(x), ensemble_size])

    for i in range(0,ensemble_size):
        case_id[i], sol[:,i] = analytical(E[i],nu[i],L,W,H,d,p[i],x)

    return case_id, sol    

def main():
    try:
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
        print "info: MPI found"
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
        print "info: MPI not found => MPI disabled"


    from fwk.wutils import parseargs
    args=parseargs()

    from PyTrilinos import Teuchos
    import numpy as np
    import matplotlib.pyplot as plt

    import mrstlnos.utilities as u


    DOFperNode = 3

    display = True

    ## Mesh

    geo_name = 'beam.geo'
    mesh_name = 'beam.msh'
    partitioned_mesh_name = 'beam_k.msh'

    par={}
    par['L'] = 50.
    par['W'] = 5.
    par['H'] = 5.
    par['d'] = 0.
    par['n_rL'] = 14
    par['n_hW'] = 3
    par['n_hH'] = 3

    ensemble_size = 1

    import shlex, subprocess
    import os

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    pars={ 'L' : par['L'] , 'W' : par['W'], 'H' : par['H'], 'd' : par['d'], 'n_rL' : par['n_rL'], 'n_hW' : par['n_hW'], 'n_hH' : par['n_hH']}
    gmsh.MeshLoader(geo_name,__file__).execute(**pars)

    if rank == 0:
        command_line = 'gmsh -3 ' + work_dir + '/' +  mesh_name + ' -o ' + work_dir + '/' +  partitioned_mesh_name +' -part ' + str(siz)
        print command_line
        tmp = shlex.split(command_line)
        fileout = open('gmsh.log','w')
        p = subprocess.Popen(tmp, stdin=subprocess.PIPE, stdout=fileout, stderr=fileout, env=os.environ, shell=False,close_fds=True)
        retcode = p.wait()
        for i in range(1,siz):
            comm.send(1, dest=i, tag=11)
    else:
        data = comm.recv(source=0, tag=11)

    if siz == 1:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).execute()
    else:
        msh = gmsh.MeshLoader(partitioned_mesh_name,work_dir).myrank_execute(rank)

    for n1 in msh.nodes:
        n1.row = n1.no-1

    ## Problem

    #msh = gmsh.MeshLoader(mesh_name,work_dir).execute()        



    pbl = m.Problem(msh,comm)

    
    E = 21000 * np.ones(ensemble_size)
    nu = 0.29 * np.ones(ensemble_size)
    k = 0
    d = 0

    #for i in range(0,ensemble_size):
    #    E[i] = 21000+i*5000

    m.Medium(pbl, "Beam", "Mat 1",E,nu,k,d)
    m.Medium(pbl, "Rigid body", "Mat 1",E,nu,k,d)

    dx = np.zeros(ensemble_size)
    dy = np.zeros(ensemble_size)
    dz = 0*np.linspace(0,-2.,ensemble_size)
    dz2 = np.zeros(ensemble_size)
    dT = np.zeros(ensemble_size)
    m.uDirichlet(pbl, "Rigid clamped surface", "Clamped 1",1,dx,1,dy,1,dz,0,dT)
    #m.uDirichlet(pbl, "Rigid clamped surface", "Clamped 1",1,0.,1,0.,1,0.,1,0.,ensemble_size)
    #m.uDirichlet(pbl, "Rigid clamped surface", "Clamped 1",1,0.,1,0.,1,0.,1,0.,ensemble_size)
    m.uDirichlet(pbl, "Clamped end", "Clamped 1",1,dx,1,dy,1,dz2,0,dT)
    m.uDirichlet(pbl, "Free end", "Clamped 1",1,0.,0,0.,0,0.,0,0.,ensemble_size)

    #p = 0.82* np.ones(ensemble_size)
    
    #p = np.linspace(0.1, 0.835, num=ensemble_size) 
    p = 0.9 * np.ones(ensemble_size) #0*np.linspace(1., 20., num=ensemble_size) 
    for i in range(0,ensemble_size):
        p[i] = 0.5+i*0.025

    zero = np.zeros(ensemble_size)

    m.Neumann(pbl, "Loaded surface", "Load 1",0,zero,0,zero,1,-p,0,zero,ensemble_size)
    m.Neumann(pbl, "Loaded surface 2", "Load 2",0,zero,0,zero,1,-p,0,zero,ensemble_size)


    norm1 = tbox.Pt(0,0,-1)
    norm2 = tbox.Pt(0,0,1)
    dg = 0*np.linspace(0., -2., num=ensemble_size)

    cont1 = m.Contact(pbl,"Beam contact zone","contact",norm1,dg)
    #cont1.setSlaveBoundary(work_dir+"/"+mesh_name, "Beam contact boundary")
    #cont1.setMasterBoundary(work_dir+"/"+mesh_name, "Rigid surface boundary")
    cont1.setInitialyOpen()
    #cont1.setDualShapeFunctions()
    cont1.setSignoriniContact()
    cont1.setMaster(pbl,"Rigid surface",norm2)


    ## Solver

    precondList = Teuchos.ParameterList()
    precondList = Teuchos.ParameterList()
    precondSubList = Teuchos.ParameterList()
    precondSubSubList = Teuchos.ParameterList()
    precondSubSubList['fact: iluk level-of-fill'] = 2
    precondSubList['Prec Type'] = "RILUK"
    precondSubList['Overlap'] = 1
    precondSubList['Ifpack2 Settings'] = precondSubSubList
    precondList['Ifpack2'] = precondSubList

    mueluParams = Teuchos.ParameterList()
    mueluParams['number of equations'] = DOFperNode
    mueluParams['max levels'] = 6
    mueluParams['multigrid algorithm'] = "sa"
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()

    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 4
    #mueluParamsSub2['relaxation: damping factor'] = 1.25

    mueluParams['coarse: type'] = "RELAXATION"

    mueluParamsSub3 = Teuchos.ParameterList()

    mueluParamsSub3['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub3['relaxation: sweeps'] = 4
    #mueluParamsSub3['relaxation: damping factor'] = 1.25

    mueluParams['smoother: params'] = mueluParamsSub2
    mueluParams['coarse: params'] = mueluParamsSub3

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 100000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 100
    solverList['ifpackParams'] = precondList
    solverList['mueluParams'] = mueluParams

    #solverList["MueLu xml file name"] = file_dir + '/finest_level_prec_simple.xml'
    solverList["MueLu xml file name"] = file_dir + '/nested_multigrid_2_lvls.xml'
    solverList["convert MueLu xml file"] = True

    solverList['Use preconditioner'] = True

    solverList['type']= "BlockGmres"
    solverList['Maximum active set iteration'] = 3
    slv = m.IterativeSolver(pbl,args.k,solverList,DOFperNode, ensemble_size>1, ensemble_size)

    #solverList['type']= "KLU2"
    #slv = m.DirectSolver(pbl,args.k,solverList,DOFperNode, ensemble_size>1, ensemble_size)

    #slv = m.Solver(pbl,args.k,solverList,DOFperNode)


    #tag_names = ["Clamped end","Free end","Rigid clamped surface"]
    #u.print_nodes(msh, tag_names, rank)

    ## Post process mesh
    if siz > 1:
        msh2 = gmsh.MeshLoader(partitioned_mesh_name,work_dir).execute()
        for n1 in msh2.nodes:
            n1.row = n1.no-1

    if not args.nogui:
        if rank == 0:
            import mrstlnos.viewer as v
            elev = -90.0
            azim = 0.0
            if siz > 1:
                gui = v.MeshViewer(pbl,slv,msh2,elev,azim)
            else:
                gui = v.MeshViewer(pbl,slv,msh,elev,azim)
            gui.vmin = 0
            gui.vmax = 0.1
            gui.start()
            gui.writeXML()
        else:
            slv.start()
    else:
        slv.start()

    '''
    if rank == 0:
        print slv.tms['belos'].read().getReal().sec()
        slv.tms['belos'].reset()
        print slv.tms['belos'].read().getReal().sec()
        slv.tms.reset() 
        print slv.tms   
    '''

    ## Extract results
    if rank == 0:
        tag_name = "Center of cross section"
        which_pos = 0
        which_dof = 2

        pos, sol = u.extract(msh, tag_name, DOFperNode, which_pos, which_dof, ensemble_size,"solution.text")

        ## Analytical results

        L = par['L']
        W = par['W']
        H = par['H']
        d = par['d']

        case_id, sol_ana = analytical_ensemble(E,nu,L,W,H,d,p,pos,ensemble_size)

        #print sol[len(sol[:,0])-1,0]
        #print sol_ana[len(sol_ana[:,0])-1,0]

        from fwk.testing import *
        tests = CTests()
        tests.add(CTest('faked measure', 0, 0))
        tests.run()

        colors = {0:(0,0.4470,0.7410), 1:(0.8500,0.3250,0.0980), 2:(0.9290,0.6940,0.1250), 3:(0.4940,0.1840,0.5560), 4:(0.4660,0.6740,0.1880), 5:(0.3010,0.7450,0.9330), 6:(0.6350,0.0780,0.1840), 7:(0,0.4470,0.7410)}

        if display:
            fig, ax = plt.subplots(3, sharex=True, figsize=(15, 8))
            for i in range(0,ensemble_size):
                ax[0].plot(pos, sol[:,i],color=colors[i%8], linewidth=2.0)
                ax[0].plot(pos, sol_ana[:,i],'--',color=colors[i%8], linewidth=2.0)
                ax[1].plot(pos, sol[:,i]-sol_ana[:,i],color=colors[i%8], linewidth=2.0)
                ax[2].plot(pos, (sol[:,i]-sol_ana[:,i])/(sol_ana[len(sol_ana[:,i])-1,i]),color=colors[i%8], linewidth=2.0)

            ax[0].set_title('Case 1: Solution along '+ tag_name)

            plt.xlabel('x [cm]')
            ax[0].set_ylabel('w')
            ax[0].set_ylim([-1.2*d, 0.])

            ax[1].set_ylabel('d w')
            ax[2].set_ylabel('d w/w(L)')

            ax[0].grid(True)
            ax[1].grid(True)
            ax[2].grid(True)
            plt.savefig("deflection.eps", format='eps', dpi=1200)
            plt.show()

if __name__ == "__main__":
    main()
