//@HEADER
// ************************************************************************
//
//                 Belos: Block Linear Solvers Package
//                  Copyright 2004 Sandia Corporation
//
// Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
// the U.S. Government retains certain rights in this software.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the Corporation nor the names of the
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact Michael A. Heroux (maherou@sandia.gov)
//
// ************************************************************************
//@HEADER

#ifndef BELOS_STATUS_TEST_GEN_IMPRESNORM_MP_VECTOR_DECL_H
#define BELOS_STATUS_TEST_GEN_IMPRESNORM_MP_VECTOR_DECL_H

/*!
  \file Belos_StatusTest_ImpResNorm_MP_Vector.hpp
  \brief Belos::StatusTest for specifying an implicit residual norm stopping criteria that checks for loss of accuracy.

  Specialized for Sacado::MP::Vector scalar type to track how many iterations
  each component of ensemble takes.
*/

#include "BelosStatusTestResNorm_MP_Vector.hpp"
#include "Stokhos_Sacado_Kokkos_MP_Vector.hpp"
#include "BelosStatusTestImpResNorm.hpp"

#include "EnsembleTraits.h"

/*!
  \class Belos::StatusTestImpResNorm
  \brief Convergence test using the implicit residual norm(s), with an
    explicit residual norm(s) check for loss of accuracy if necessary.

  This test passes when a quorum of residual vectors (for all
  right-hand sides) passes a residual norm test.  For residual vector
  \f$r_i\f$, the form of the test is \f$\|r_i\| / \sigma_i \le
  \tau\f$, where \f$\tau\f$ is the convergence tolerance set in the
  constructor or by setTolerance().  The default quorum consists of
  all the vectors, but you can set the number of vectors that must
  pass before the whole test passes, either by the constructor
  argument or by calling setQuorum().

  The default residual vector norm (the norm used in the numerator of
  the test, and applied to the current residual vectors) is the
  2-norm, but you can change the norm type (1-norm, 2-norm, or
  infinity norm) using defineResForm().  The default scaling factor
  \f$\sigma_i\f$ for residual vector i is the 2-norm of the initial
  residual vector, but you can change this using defineScaleForm().
  The norm used for the scaling factors may be different than the norm
  used for the current residual vectors.  Furthermore, each residual
  vector may use a different scaling factor \f$\sigma_i\f$.

  This test starts by using the "implicit" residual norm, otherwise
  called the "recursive" or "native" residual norm.  It comes from the
  iterative method's projected problem, unlike the "explicit" residual
  norm, which comes from explicitly computing \f$R = B - A * X\f$.
  Once the implicit residual norm reaches the convergence tolerance,
  this test then checks the explicit residual norm to make sure that
  it has reached the convergence tolerance as well.

  We say that there is a "potential loss of accuracy" when we first
  detect that the implicit residual norm(s) have met the desired
  original convergence tolerance, but the explicit residual norm(s)
  have not.  We don't actually call this a "loss of accuracy" (in the
  sense of getLOADetected() returning true) unless we tried to remedy
  this situation, but couldn't fix it.  Upon detecting a potential
  loss of accuracy, this test tells the solver to "iterate a few more
  steps" by making the "current" residual tolerance (the value of
  getCurrTolerance()) smaller, and forcing the implicit residual
  norm(s) to meet the current tolerance before moving on to test the
  explicit norm(s).  If that doesn't work, this test declares a loss
  of accuracy.

   This implementation is the same as the default, but is specialized for
   the Sacado::MP::Vector scalar type to keep track of when each component
   of the ensemble converges.
  
*/

namespace Belos {

template <class Storage, class MV, class OP>
class StatusTestImpResNorm<Sacado::MP::Vector<Storage>, MV, OP> :
    public StatusTestResNorm<Sacado::MP::Vector<Storage>,MV,OP> {

public:
  typedef Sacado::MP::Vector<Storage> ScalarType;
  //! The type of the magnitude (absolute value) of a ScalarType.
  typedef typename Teuchos::ScalarTraits<ScalarType>::magnitudeType MagnitudeType;

private:
  //! @name Abbreviations for method implementations
  //@{
  typedef Teuchos::ScalarTraits<ScalarType> STS;
  typedef Teuchos::ScalarTraits<MagnitudeType> STM;
  typedef MultiVecTraits<ScalarType,MV> MVT;
  //@}

public:
  //! @name Constructors and destructor.
  //@{

  /// \brief Constructor
  ///
  /// \param Tolerance [in] Convergence tolerance \f$\tau\f$.
  ///
  /// \param quorum [in] The number of vectors in the problem that
  ///   must pass the convergence criteria before this StatusTest
  ///   passes.  -1 means that all residuals must pass.
  ///
  /// \param showMaxResNormOnly [in] Whether only the maximum residual
  ///   norm (of all vectors) is displayed when the print() method is
  ///   called.
  StatusTestImpResNorm (MagnitudeType Tolerance,
                        int quorum = -1,
                        bool showMaxResNormOnly = false);

  //! Destructor (virtual for memory safety).
  virtual ~StatusTestImpResNorm();

  //@}
  //! @name Form and parameter definition methods.
  //@{

  //! Define form of the residual, its norm and optional weighting vector.
  /*! This method defines the form of \f$\|r\|\f$.  We specify:
    <ul>
    <li> The norm to be used on the residual (this may be different than the norm used in
    defineScaleForm()).
    </ul>
  */
  int defineResForm( NormType TypeOfNorm);

  //! Define form of the scaling, its norm, its optional weighting vector, or, alternatively, define an explicit value.
  /*! This method defines the form of how the residual is scaled (if at all).  It operates in two modes:
    <ol>
    <li> User-provided scaling value:
    <ul>
    <li> Set argument TypeOfScaling to UserProvided.
    <li> Set ScaleValue to a non-zero value that the residual norm will be divided by.
    <li> TypeOfNorm argument will be ignored.
    <li> Sample use:  Define ScaleValue = \f$\|A\|_{\infty}\f$ where \f$ A \f$ is the matrix
    of the linear problem.
    </ul>

    <li> Use a supported Scaling Form:
    <ul>
    <li> Define TypeOfScaling to be the norm of the right hand side, the initial residual vector,
    or to none.
    <li> Define norm to be used on the scaling vector (this may be different than the norm used
    in DefineResForm()).
    </ul>
    </ol>
  */
  int defineScaleForm( ScaleType TypeOfScaling, NormType TypeOfNorm, MagnitudeType ScaleValue = Teuchos::ScalarTraits<MagnitudeType>::one());

  //! Set the value of the tolerance
  /*! We allow the tolerance to be reset for cases where, in the process of testing the residual,
    we find that the initial tolerance was too tight or too lax.
  */
  int setTolerance (MagnitudeType tolerance) {
    tolerance_ = tolerance;
    return 0;
  }

  //! Sets the number of residuals that must pass the convergence test before Passed is returned.
  //! \note If \c quorum=-1 then all residuals must pass the convergence test before Passed is returned.
  int setQuorum (int quorum) {
    quorum_ = quorum;
    return 0;
  }

  //! Set whether the only maximum residual norm is displayed when the print() method is called
  int setShowMaxResNormOnly (bool showMaxResNormOnly) {
    showMaxResNormOnly_ = showMaxResNormOnly;
    return 0;
  }

  //@}

  //! @name Status methods
  //@{
  //! Check convergence status: Passed, Failed, or Undefined.
  /*! This method checks to see if the convergence criteria are met.
    Depending on how the residual test is constructed this method will return
    the appropriate status type.

    \return StatusType: Passed, Failed, or Undefined.
  */
  StatusType checkStatus(Iteration<ScalarType,MV,OP>* iSolver);

  //! Return the result of the most recent CheckStatus call.
  StatusType getStatus() const {return(status_);};
  //@}

  //! @name Reset methods
  //@{

  //! Resets the internal configuration to the initial state.
  void reset();

  //@}

  //! @name Print methods
  //@{

  //! Output formatted description of stopping test to output stream.
  void print(std::ostream& os, int indent = 0) const;

  //! Print message for each status specific to this stopping test.
  void printStatus(std::ostream& os, StatusType type) const;
  //@}

  //! @name Methods to access data members.
  //@{

  //! Returns the current solution estimate that was computed for the most recent residual test.
  Teuchos::RCP<MV> getSolution() { return curSoln_; }

  //! Returns the number of residuals that must pass the convergence test before Passed is returned.
  //! \note If \c quorum=-1 then all residuals must pass the convergence test before Passed is returned.
  int getQuorum() const { return quorum_; }

  //! Returns whether the only maximum residual norm is displayed when the print() method is called
  bool getShowMaxResNormOnly() { return showMaxResNormOnly_; }

  //! Returns the vector containing the indices of the residuals that passed the test.
  std::vector<int> convIndices() { return ind_; }

  /// \brief "Original" convergence tolerance \f$\tau\f$ as set by user.
  ///
  /// This value is the convergence tolerance as set by the user in
  /// the constructor, or by the setTolerance() method.  See this
  /// class' main documentation and the documentation of
  /// getCurrTolerance() for an explanation of the difference between
  /// the "original" and "current" tolerances.
  MagnitudeType getTolerance () const {
    return tolerance_;
  }

  /// \brief Current convergence tolerance; may be changed to prevent loss of accuracy.
  ///
  /// The difference between "original" tolerance (the value of
  /// getTolerance()) and "current" tolerance (the value of this
  /// method) relates to the idea of "loss of accuracy."  See this
  /// class' main documentation for details.  We do not allow users to
  /// set the "current" tolerance.
  MagnitudeType getCurrTolerance () const {
    return currTolerance_;
  }

  //! Returns the test value, \f$ \frac{\|r\|}{\sigma} \f$, computed in most recent call to CheckStatus.
  const std::vector<MagnitudeType>* getTestValue() const {return(&testvector_);};

  //! Returns the residual norm value, \f$ \|r\| \f$, computed in most recent call to CheckStatus.
  const std::vector<MagnitudeType>* getResNormValue() const {return(&resvector_);};

  //! Returns the scaled norm value, \f$ \sigma \f$.
  const std::vector<MagnitudeType>* getScaledNormValue() const {return(&scalevector_);};

  //! Returns a boolean indicating a loss of accuracy has been detected in computing the residual.
  bool getLOADetected() const { return lossDetected_; }

  //! Returns number of ensemble iterations
  const std::vector<int> getEnsembleIterations() const { return ensemble_iterations; }
  //@}


  /** @name Misc. */
  //@{

  /** \brief Call to setup initial scaling vector.
    *
    * After this function is called <tt>getScaledNormValue()</tt> can be called
    * to get the scaling vector.
    */
  StatusType firstCallCheckStatusSetup(Iteration<ScalarType,MV,OP>* iSolver);
  //@}

  /** \name Overridden from Teuchos::Describable */
  //@{

  /** \brief Method to return description of the maximum iteration status test  */
  std::string description() const
  {
    std::ostringstream oss;
    oss << "Belos::StatusTestImpResNorm<>: " << resFormStr();
    oss << ", tol = " << tolerance_;
    return oss.str();
  }
  //@}

  protected:

  private:

  //! @name Private methods.
  //@{
  /** \brief Description of current residual form */
  std::string resFormStr() const
  {
    std::ostringstream oss;
    oss << "(";
    oss << ((resnormtype_==OneNorm) ? "1-Norm" : (resnormtype_==TwoNorm) ? "2-Norm" : "Inf-Norm");
    oss << " Res Vec) ";

    // If there is no residual scaling, return current string.
    if (scaletype_!=None)
    {
      // Insert division sign.
      oss << "/ ";

      // Determine output string for scaling, if there is any.
      if (scaletype_==UserProvided)
        oss << " (User Scale)";
      else {
        oss << "(";
        oss << ((scalenormtype_==OneNorm) ? "1-Norm" : (resnormtype_==TwoNorm) ? "2-Norm" : "Inf-Norm");
        if (scaletype_==NormOfInitRes)
          oss << " Res0";
        else if (scaletype_==NormOfPrecInitRes)
          oss << " Prec Res0";
        else
          oss << " RHS ";
        oss << ")";
      }
    }

    return oss.str();
  }

  //! @name Private data members.
  //@{

  //! Current tolerance used to determine convergence and the default tolerance set by user.
  MagnitudeType tolerance_, currTolerance_;

  //! Number of residuals that must pass the convergence test before Passed is returned.
  int quorum_;

  //! Determines if the entries for all of the residuals are shown or just the max.
  bool showMaxResNormOnly_;

  //! Type of norm to use on residual (OneNorm, TwoNorm, or InfNorm).
  NormType resnormtype_;

  //! Type of scaling to use (Norm of RHS, Norm of Initial Residual, None or User provided)
  ScaleType scaletype_;

  //! Type of norm to use on the scaling (OneNorm, TwoNorm, or InfNorm)
  NormType scalenormtype_;

  //! Scaling value.
  MagnitudeType scalevalue_;

  //! Scaling vector.
  std::vector<MagnitudeType> scalevector_;

  //! Residual norm vector.
  std::vector<MagnitudeType> resvector_;

  //! Test vector = resvector_ / scalevector_
  std::vector<MagnitudeType> testvector_;

  //! Current solution vector
  Teuchos::RCP<MV> curSoln_;

  //! Vector containing the indices for the vectors that passed the test.
  std::vector<int> ind_;

  //! Status
  StatusType status_;

  //! The current blocksize of the linear system being solved.
  int curBlksz_;

  //! The current number of right-hand sides being solved for.
  int curNumRHS_;

  //! The indices of the current number of right-hand sides being solved for.
  std::vector<int> curLSIdx_;

  //! The current number of linear systems that have been loaded into the linear problem.
  int curLSNum_;

  //! The total number of right-hand sides being solved for.
  int numrhs_;

  //! Is this the first time CheckStatus is called?
  bool firstcallCheckStatus_;

  //! Is this the first time DefineResForm is called?
  bool firstcallDefineResForm_;

  //! Is this the first time DefineScaleForm is called?
  bool firstcallDefineScaleForm_;

  //! Has a loss in accuracy been detected?
  bool lossDetected_;

  //! The number of iterations at which point each ensemble component converges
  std::vector<int> ensemble_iterations;

  //@}

};


} // end namespace Belos

#endif /* BELOS_STATUS_TEST_GEN_IMPRESNORM_MP_VECTOR_DECL_H */
    
