#ifndef STOKHOS_MP_VECTOR_MASKTRAITS_H
#define STOKHOS_MP_VECTOR_MASKTRAITS_H

#include "Sacado.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include <iostream>

template <typename T>
struct EnsembleTraits_m {
  static const int size = 1;
  typedef T value_type;
  static const value_type& coeff(const T& x, int i) { return x; }
  static value_type& coeff(T& x, int i) { return x; }
};

template <typename S>
struct EnsembleTraits_m< Sacado::MP::Vector<S> > {
  static const int size = S::static_size;
  typedef typename S::value_type value_type;
  static const value_type& coeff(const Sacado::MP::Vector<S>& x, int i) {
    return x.fastAccessCoeff(i);
  }
  static value_type& coeff(Sacado::MP::Vector<S>& x, int i) {
    return x.fastAccessCoeff(i);
  }
};

template<typename scalar> class Mask 
{
    private: 
        static const int size = EnsembleTraits_m<scalar>::size;
        bool data[size];

    public:
        Mask(){
        }

        Mask(bool a){
            size = 1;
            data = new bool[size]();
            data[0]=a;
        }

        int getSize() const {return size;}

        bool operator> (double v)
        {
            double sum = 0;
            for(auto i=0; i<size; ++i)
                sum = sum + this->data[i];

            return sum > v*size;
        }

        bool operator< (double v)
        {
            double sum = 0;
            for(auto i=0; i<size; ++i)
                sum = sum + this->data[i];

            return sum < v*size;
        }        

        bool operator== (double v)
        {
            double sum = 0;
            for(auto i=0; i<size; ++i)
                sum = sum + this->data[i];

            return sum == v*size;
        } 

        bool operator!= (double v)
        {
            double sum = 0;
            for(auto i=0; i<size; ++i)
                sum = sum + this->data[i];

            return sum != v*size;
        }

        Mask<scalar> operator&& (Mask<scalar> m2)
        {
            Mask<scalar> m3;
            for(auto i=0; i<size; ++i)
                m3.data[i] = (this->data[i] && m2.data[i]);

            return m3;
        } 

        Mask<scalar> operator|| (Mask<scalar> m2)
        {
            Mask<scalar> m3;
            for(auto i=0; i<size; ++i)
                m3.data[i] = (this->data[i] || m2.data[i]);

            return m3;
        } 

        Mask<scalar> operator+ (Mask<scalar> m2)
        {
            Mask<scalar> m3;
            for(auto i=0; i<size; ++i)
                m3.data[i] = (this->data[i] + m2.data[i]);

            return m3;
        } 

        Mask<scalar> operator- (Mask<scalar> m2)
        {
            Mask<scalar> m3;
            for(auto i=0; i<size; ++i)
                m3.data[i] = (this->data[i] - m2.data[i]);

            return m3;
        }

        scalar operator* (scalar v)       
        {
            typedef EnsembleTraits_m<scalar> ET; 
            scalar v2;
            for(auto i=0; i<size; ++i)
                ET::coeff(v2,i) = ET::coeff(v,i)*this->data[i];

            return v2;            
        }

        bool operator[] (int i) const
        {
            return this->data[i];
        }

        bool & operator[] (int i)
        {
            return this->data[i];
        }

        Mask<scalar> operator! ()
        {
            Mask<scalar> m2;
            for(auto i=0; i<size; ++i)
                m2.data[i] = !(this->data[i]);

            return m2;            
        }

        operator bool() const
        {
            return this->data[0];                            
        }   

        operator double() const
        {
            double sum = 0;
            for(auto i=0; i<size; ++i)
                sum = sum + this->data[i];

            return sum/size;    
        }                               
} ;

template<typename scalar> std::ostream &operator<<(std::ostream &os, const Mask<scalar>& m) { 
    os << "[ "; 
    for(int i=0; i<m.getSize(); ++i)
        os << m[i] << " ";
    return os << "]";    
}

template<typename S> Mask<Sacado::MP::Vector<S> > operator> (Sacado::MP::Vector<S> &a1, Sacado::MP::Vector<S> &a2)
{   
    typedef EnsembleTraits_m<Sacado::MP::Vector<S> > ET;

    Mask<Sacado::MP::Vector<S> > mask;
    for(auto i=0; i<ET::size; ++i)
        mask[i] = ET::coeff(a1,i) > ET::coeff(a2,i);
    return mask;
};

template<typename S> Mask<Sacado::MP::Vector<S> > operator>= (Sacado::MP::Vector<S> &a1, Sacado::MP::Vector<S> &a2)
{   
    typedef EnsembleTraits_m<Sacado::MP::Vector<S> > ET;

    Mask<Sacado::MP::Vector<S> > mask;
    for(auto i=0; i<ET::size; ++i)
        mask[i] = ET::coeff(a1,i) >= ET::coeff(a2,i);
    return mask;
};

template<typename S> Mask<Sacado::MP::Vector<S> > operator< (Sacado::MP::Vector<S> &a1, Sacado::MP::Vector<S> &a2)
{   
    typedef EnsembleTraits_m<Sacado::MP::Vector<S> > ET;

    Mask<Sacado::MP::Vector<S> > mask;
    for(auto i=0; i<ET::size; ++i)
        mask[i] = ET::coeff(a1,i) < ET::coeff(a2,i);
    return mask;
};

template<typename S> Mask<Sacado::MP::Vector<S> > operator<= (Sacado::MP::Vector<S> &a1, Sacado::MP::Vector<S> &a2)
{   
    typedef EnsembleTraits_m<Sacado::MP::Vector<S> > ET;

    Mask<Sacado::MP::Vector<S> > mask;
    for(auto i=0; i<ET::size; ++i)
        mask[i] = ET::coeff(a1,i) <= ET::coeff(a2,i);
    return mask;
};

namespace MaskLogic{

class OR
{
private:
    bool value;
    
public:
    OR(){
        value = false;
    }

    OR(bool value_){
        value = value_;
    }

    template<typename T> OR(T m){
        value = (((double) m)!=0.);
    }
    
    operator bool() const
    {
        return value;
    }
};

class XOR
{
private:
    bool value;
    
public:
    XOR(){
        value = false;
    }

    XOR(bool value_){
        value = value_;
    }

    template<typename T> XOR(T m){
        value = (((double) m)==1./m.getSize());
    }
    
    operator bool() const
    {
        return value;
    }
};

class AND
{
private:
    bool value;
    
public:
    AND(){
        value = false;
    }

    AND(bool value_){
        value = value_;
    }
   
    template<typename T> AND(T m){
        value = (((double) m)==1.);
    }
    
    operator bool() const
    {
        return value;
    }
};
}
#endif // STOKHOS_MP_VECTOR_MASKTRAITS_H
