#ifndef WNEUMANN_LIST_H
#define WNEUMANN_LIST_H

#include "mrstlnos.h"

#include "Map.h"
#include "wNeumann.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

#include "EnsembleTraits.h"

using Teuchos::RCP;

namespace mrstlnos
{
    namespace discrete_problem
    {
        template<typename scalar> class Neumann_list
        {
        private:
            Kokkos::View<int**, Kokkos::LayoutRight>  neumann_elements_list;
            Kokkos::View<scalar***, Kokkos::LayoutRight>  neumann_values_list;

            Tpetra::Vector<>::local_ordinal_type neumann_number;
            Tpetra::Vector<>::local_ordinal_type neumann_size;
            Tpetra::Vector<>::local_ordinal_type neumann_elements;

            Tpetra::Vector<>::local_ordinal_type neumann_time;
            Tpetra::Vector<>::local_ordinal_type neumann_dofs;
            Tpetra::Vector<>::local_ordinal_type neumann_values;
        public:
            Neumann_list();
            Neumann_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list);
            inline local_ordinal_type getNeumannNumber(){return neumann_number;};
            inline local_ordinal_type getNeumannSize(local_ordinal_type i)
                    {return neumann_elements_list(i,neumann_size);};
            inline local_ordinal_type getNeumannElement(local_ordinal_type i, local_ordinal_type n1)
                    {return neumann_elements_list(i,neumann_elements + n1);};
            inline scalar getNeumannDof(local_ordinal_type i, local_ordinal_type j)
                    {return neumann_values_list(i,0,neumann_dofs+j);};
            inline scalar getNeumannValue(local_ordinal_type i, local_ordinal_type j)
                    {return neumann_values_list(i,0,neumann_values+j);};
        };

        template<typename scalar> Neumann_list<scalar>::Neumann_list()
        {
            neumann_number = 0;
            neumann_size = 0;
            neumann_elements = 0;

            neumann_time = 0;
            neumann_dofs = 0;
            neumann_values = 0;

            neumann_elements_list = Kokkos::View<int**, Kokkos::LayoutRight> ("R",neumann_number,0);
            neumann_values_list = Kokkos::View<scalar***, Kokkos::LayoutRight> ("R",neumann_number,0,0);

        }

        template<typename scalar> Neumann_list<scalar>::Neumann_list(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<Elements_list> elements_list)
        {
            typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
            typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

            typedef EnsembleTraits<scalar> ET;
            const int ensemble_size = ET::size;

            local_ordinal_type numDPN = map->numPrimalDPN;

            neumann_number = pbl.nBCs.size();
            local_ordinal_type maxNumElemsPernBC = 0;

            std::vector<std::vector<global_ordinal_type>> nbcElems ={};

            for (auto i = 0; i < neumann_number; ++i)
            {
                std::vector<global_ordinal_type> mynBcElems ={};
                for (auto j = 0; j < pbl.nBCs[i]->tag->elems.size(); ++j)
                {
                    local_ordinal_type e = map->mapElems->getLocalElement(pbl.nBCs[i]->tag->elems[j]->no-1);

                    if      (elements_list->Get_Element_Type(e) == static_cast<int>(ELTYPE::QUAD4));
                    else if (elements_list->Get_Element_Type(e) == static_cast<int>(ELTYPE::TRI3) );
                    else
                        continue;

                    mynBcElems.push_back(e);

                }
                nbcElems.push_back(mynBcElems);
                if (maxNumElemsPernBC < mynBcElems.size())
                    maxNumElemsPernBC = mynBcElems.size();
            }

            neumann_size = 0;
            neumann_elements = 1;
            neumann_elements_list = Kokkos::View<int**, Kokkos::LayoutRight> ("R",neumann_number,neumann_elements+maxNumElemsPernBC);

            for (auto i = 0; i < neumann_number; ++i)
            {
                neumann_elements_list(i,neumann_size) = nbcElems[i].size();

                for ( auto j = 0; j < nbcElems[i].size() ; ++j)
                    neumann_elements_list(i,neumann_elements+j) = nbcElems[i][j];
            }

            neumann_time = 0;
            neumann_dofs = 1;
            neumann_values = 1+numDPN;
            neumann_values_list = Kokkos::View<scalar***, Kokkos::LayoutRight> ("R",neumann_number,1,1+2*numDPN);
            for (auto i = 0; i < neumann_number; ++i)
            {
                neumann_values_list(i,0,neumann_time) = 0.; //to change latter
                for ( auto j = 0; j < numDPN ; ++j)
                    if(pbl.nBCs[i]->which_dof[j])
                        neumann_values_list(i,0,neumann_dofs+j) = 1;
                    else
                        neumann_values_list(i,0,neumann_dofs+j) = 0;

                for ( auto j = 0; j < numDPN ; ++j)
                for (int s=0; s<ensemble_size; ++s)
                {
                    ET::coeff(neumann_values_list(i,0,neumann_values+j),s) = pbl.nBCs[i]->values[j*ensemble_size+s];
                }

                //for ( auto j = 0; j < numDPN ; ++j)
                //    neumann_values_list(i,0,neumann_values+j) = pbl.nBCs[i]->values[j];

                //for ( auto j = 0; j < numDPN ; ++j)
                //    {
                //        std::cout << i << ", " << j << " " << pbl.nBCs[i]->which_dof[j] << " " << neumann_values_list(i,0,neumann_dofs+j) << " "<< this->getNeumannDof(i,j) << " " << neumann_values_list(i,0,neumann_values+j) << " "<< pbl.nBCs[i]->values[j] << " "<<this->getNeumannValue(i,j) << std::endl;
                //    }
            }

        }
    };
};

#endif //WNEUMANN_LIST_H
