#include "wHex8.h"
#include "wSfHex8.h"
#include "wGaussHex8.h"
#include "wCacheHex8.h"
#include "wMemHex8.h"

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"

/**
 * @brief Those functions compute the stress and strain at nodes and in average per element.
 * They are currently implemented using gmm data structure but should be refactored using Kokkos:view
 * such that it will be possible to use parallel for in the future.
 * @authors Kim Liegeois
 */

void strain_at_gp(size_t n_n, Mem &mem, Cache &cache, size_t a, gmm::dense_matrix<double> u_e, gmm::dense_matrix<double> &Strain)
{
    gmm::dense_matrix<double> const &J = mem.getJinv(a);

    gmm::dense_matrix<double> du_d_Phi(3, 3);
    gmm::dense_matrix<double> Tmp(3, 3);
    gmm::dense_matrix<double> Tmp2(3, 3);

    for (size_t i = 0; i < n_n; ++i)
    {
        std::vector<double> &dffi = cache.dff[a][i];
        for (auto k = 0; k < 3; ++k)
            for (auto j = 0; j < 3; ++j)
                du_d_Phi(j, k) += dffi[j] * u_e(k, i);
    }

    gmm::mult(J, du_d_Phi, Tmp);
    gmm::add(gmm::transposed(Tmp), Tmp, Tmp2);
    gmm::copy(gmm::scaled(Tmp2, 0.5), Tmp);
    gmm::add(Tmp, Strain);
}

void compute_stress(size_t n_n, gmm::dense_matrix<double> Strain, gmm::dense_matrix<double> &Stress, const gmm::dense_matrix<double> &H, const gmm::dense_matrix<double> &D, const std::vector<double> &T, bool thermal)
{
    gmm::clear(Stress);

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);

    if (thermal)
        for (size_t i = 0; i < 3; ++i)
            for (size_t j = 0; j < 3; ++j)
            {
                double tmp = 0.;
                for (size_t k = 0; k < n_n; ++k)
                    tmp += D(i, j) * T[k];
                Stress(i, j) += tmp;
            }
}

void strain_stress_x_y_z(Element *elem, gmm::dense_matrix<double> &Strain, gmm::dense_matrix<double> &Stress, const gmm::dense_matrix<double> &H, const gmm::dense_matrix<double> &D, const std::vector<double> &x, const std::vector<double> &y, const std::vector<double> &z, const std::vector<double> &T, bool thermal = false)
{
    Cache &cache = elem->getVCache();
    Mem &mem = elem->getVMem();

    size_t n_gp = cache.dff.size();
    size_t n_n = elem->nodes.size();

    gmm::dense_matrix<double> u_e(3, n_n);
    gmm::dense_matrix<double> Tmp(3, 3);

    gmm::clear(Strain);

    for (size_t i = 0; i < n_n; ++i)
    {
        u_e(0, i) = x[i];
        u_e(1, i) = y[i];
        u_e(2, i) = z[i];
    }

    for (size_t a = 0; a < n_gp; ++a)
    {
        strain_at_gp(n_n, mem, cache, a, u_e, Strain);
    }

    gmm::copy(gmm::scaled(Strain, 1. / n_n), Tmp);
    gmm::copy(Tmp, Strain); // Prevent Level 2 warning: a conflict is possible in copy

    compute_stress(n_n, Strain, Stress, H, D, T, thermal);
}

void strain_stress_x_y_z_at_node(Element *elem, std::vector<gmm::dense_matrix<double>> &Strain_at_node, std::vector<gmm::dense_matrix<double>> &Stress_at_node, const gmm::dense_matrix<double> &H, const gmm::dense_matrix<double> &D, const std::vector<double> &x, const std::vector<double> &y, const std::vector<double> &z, const std::vector<double> &T, bool thermal = false)
{
    Cache &cache = elem->getVCache();
    Mem &mem = elem->getVMem();

    size_t n_gp = cache.dff.size();
    size_t n_n = elem->nodes.size();

    std::vector<gmm::dense_matrix<double>> Strain_at_gp(n_gp);
    std::vector<gmm::dense_matrix<double>> Stress_at_gp(n_gp);
    gmm::dense_matrix<double> u_e(3, n_n);
    gmm::dense_matrix<double> Strain(3, 3);

    for (auto &k : Strain_at_gp)
        gmm::resize(k, 3, 3);
    for (auto &k : Stress_at_gp)
        gmm::resize(k, 3, 3);

    for (size_t i = 0; i < n_n; ++i)
    {
        u_e(0, i) = x[i];
        u_e(1, i) = y[i];
        u_e(2, i) = z[i];
    }

    for (size_t a = 0; a < n_gp; ++a)
    {
        gmm::clear(Strain);
        strain_at_gp(n_n, mem, cache, a, u_e, Strain);
        gmm::copy(Strain, Strain_at_gp[a]);
        compute_stress(n_n, Strain, Stress_at_gp[a], H, D, T, thermal);
    }

    // Extrapolation at node points
    gmm::dense_matrix<double> values_at_gp(n_gp, 18);
    gmm::dense_matrix<double> values_at_node(n_n, 18);

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t a = 0; a < n_gp; ++a)
            {
                values_at_gp(a, i * 3 + j) = Strain_at_gp[a](i, j);
                values_at_gp(a, 9 + i * 3 + j) = Stress_at_gp[a](i, j);
            }

    elem->gp2node(values_at_gp, values_at_node);

    for (size_t k = 0; k < n_n; ++k)
    {
        gmm::dense_matrix<double> Strain_tmp(3, 3);
        gmm::dense_matrix<double> Stress_tmp(3, 3);
        for (size_t i = 0; i < 3; ++i)
            for (size_t j = 0; j < 3; ++j)
            {
                Strain_tmp(i, j) = values_at_node(k, i * 3 + j);
                Stress_tmp(i, j) = values_at_node(k, 9 + i * 3 + j);
            }
        gmm::copy(Strain_tmp, Strain_at_node[k]);
        gmm::copy(Stress_tmp, Stress_at_node[k]);
    }
}
