# -*- coding: utf8 -*-
# gmsh utilities

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os, os.path, sys
import fwk.wutils as wu

class MeshLoader:
    """
    Chargement d'un fichier 'geo' (execute gmsh si nécessaire)
    gmsh doit être dans le PATH
    """
    def __init__(self, filename, caller=''):
        self.filename = wu.pth(filename, caller)
        self.verb = False
            
    def convert_from_APDL(self, oldfilename):
        import tbox
        msh = tbox.MshData()
        mshConv = tbox.MshConvert(msh)
        mshConv.ansys_gmsh(oldfilename, self.filename)
        
    def execute(self, myrank=-1, **pars):
        print " [%s] processing %s" % (self.__class__, self.filename)
        ext = os.path.splitext(self.filename)[1]
        if ext == '.msh':
            return self.__loadMSH(self.filename, myrank)
        elif ext == '.geo':
            self.__execGMSH(self.filename, pars)
            return self.__loadMSH(os.path.basename(self.filename.replace('.geo','.msh')), myrank)
        else:
            raise Exception("Error %s: bad gmsh extension")

    def __buildParList(self, pars):
        parcmd=""
        for k,v in pars.iteritems():
            if k[0] == '-':
                parcmd += "%s %s " % (k,v) # option
            else:
                parcmd += "-setnumber %s %s " % (k,v) # parameter
        return parcmd

    def __execGMSH(self, geofile, pars):
        mshfile=os.path.basename(geofile.replace('.geo','.msh'))
        
        parnew=str(pars)
        if self.verb: print " [%s] new set of pars=%s" % (self.__class__, parnew)
        
        # force rebuild if pars have changed although .msh exists
        parfile=os.path.basename(geofile.replace('.geo','.par')) 
        rebuild=True  
        if os.path.isfile(mshfile):
            if self.verb: print " [%s] an older version of .msh has been found" %self.__class__ 
            if os.path.getmtime(mshfile) < os.path.getmtime(geofile):
                if self.verb: print " [%s] .geo has been modified => gmsh should be rerun" %self.__class__
            else: 
                if self.verb: print " [%s] .geo and .msh match" %self.__class__
                if os.path.isfile(parfile):
                    # a parfile exists
                    f = open(parfile,'r')
                    parold = f.readline()
                    f.close()
                    if self.verb: print " [%s] old set of pars=%s" % (self.__class__, parold)
                    if parold==parnew:
                        rebuild=False
                        if self.verb: print " [%s] pars are the same => gmsh run avoided" %self.__class__
                    else:
                        if self.verb: print " [%s] pars are different => gmsh should be rerun" %self.__class__
                else:
                    if self.verb: print " [%s] %s not found" % (self.__class__, parfile) 
        
        if rebuild:
            print ' [%s] running gmsh on "%s"...' % (self.__class__, geofile)
            parcmd = self.__buildParList(pars)
            # TEMPORARY FIX: force gmsh to produce meshes in v2 format [RB]
            cmd = 'gmsh -3 -format msh2 %s -o "%s" "%s"'% (parcmd, mshfile, geofile)
            if self.verb: print " [%s] cmd= %s" % (self.__class__, cmd)

            import shlex
            args=shlex.split(cmd)

            # lance gmsh et réagit si "gmsh" n'est pas trouvé
            import subprocess
            try:
                sproc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=os.environ)
                out, err = sproc.communicate()
            except Exception, e:
                print "Error:", e
                print "\n** Check that gmsh in your PATH!!\n"
                raise e

            # redirige la sortie de gmsh vers le stdout (eventuellement GUI) via un PIPE
            # (evite d'imprimer hors de la fenetre)
            for l in out.splitlines():
                print " [gmsh-out] ", l
            for l in err.splitlines():
                print " [gmsh-err] ", l

            if not os.path.isfile(mshfile):
                raise Exception("Problem during gmsh run")
                
            # write pars
            if self.verb: print " [%s] writing parameters to %s" % (self.__class__, parfile)
            f = open(parfile,'w')
            f.write(parnew)
            f.close()
            
        else:
            print ' [%s] "%s" up-to-date' % (self.__class__, mshfile)
        self.mshfile = mshfile
        
    def __loadMSH(self, mshfile, myrank):
        import tbox
        msh = tbox.MshData()
        gmshLoader = tbox.GmshImport(msh)
        gmshLoader.load(mshfile, myrank)
        msh.name = os.path.basename(os.path.splitext(mshfile)[0])
        return msh

# ------------------------------------------------------------------------------

 
