/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wHex8.h"
#include "wSfHex8.h"
#include "wGaussHex8.h"
#include "wCacheHex8.h"
#include "wMemHex8.h"
#include "wNode.h"
#include "wLazy.h"
#include "wFct0.h"
#include "gmm_gmm.h"

using namespace tbox;

Hex8::Hex8(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemHex8(*this, static_cast<int>(getCache().gauss.p.size()));
}
Hex8::~Hex8()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Hex8 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Hex8::getVMem()
{
    return getMem();
}
MemHex8 &Hex8::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Hex8 
 * @authors Romain Boman, Adrien Crovato
 */
CacheHex8 &Hex8::getCache()
{
    static Lazy<CacheHex8> cache(2);
    return cache.get();
}
Cache &Hex8::getVCache()
{
    return getCache();
}

/**
 * @brief Compute Jacobian Matrix at Gauss point k
 * @authors Romain Boman
 */
double Hex8::buildJ(size_t k, gmm::dense_matrix<double> &J) const
{
    CacheHex8 &cache = getCache();

    gmm::dense_matrix<double> Jn(3, 3);
    gmm::clear(J);

    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        gmm::mult(gmm::col_vector(dff), gmm::row_vector((*it)->pos.x), Jn);
        gmm::add(Jn, J);
    }
    double detJ = gmm::lu_det(J);
    return detJ;
}

void Hex8::buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(K);
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        //Pt ksi(gauss.p[k]);

        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k); // J contient maintenant Jinv!

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];

                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                K(i, j) += gmm::vect_sp(factor1, factor2) * gauss.w[k] * detJ * fk;
            }
        }
    }
}

void Hex8::buildK_mechanic(gmm::dense_matrix<double> &K_e, gmm::dense_matrix<double> &H)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(K_e);
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);
    std::vector<double> factor3(3);

    gmm::dense_matrix<double> TMP1(3, 3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));

                gmm::clear(TMP1);
                for (size_t k = 0; k < 3; ++k)
                    for (size_t m = 0; m < 3; ++m)
                    {
                        for (size_t l = 0; l < 3; ++l)
                            for (size_t n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);

                        gmm::mult(TMP1, gmm::col_vector(factor2), gmm::col_vector(factor3));
                        K_e(i * 3 + k, j * 3 + m) += gmm::vect_sp(factor1, factor3) * gauss.w[a] * detJ;
                    }
            }
        }
    }

    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = i; j < 3 * nodes.size(); ++j)
        {
            if (fabs(K_e(i, j)) < epsilon)
                K_e(i, j) = 0;
            K_e(j, i) = K_e(i, j);
        }
}

void Hex8::buildS_thermomechanic(gmm::dense_matrix<double> &S_e, gmm::dense_matrix<double> &D)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(S_e); // dimension: 24*8
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        std::vector<double> &ff = cache.ff[a];

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            gmm::mult(D, gmm::col_vector(factor1), gmm::col_vector(factor2));
            for (size_t j = 0; j < nodes.size(); ++j)
                for (size_t k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2[k] * ff[j] * gauss.w[a] * detJ;
        }
    }

    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = 0; j < nodes.size(); ++j)
            if (fabs(S_e(i, j)) < epsilon)
                S_e(i, j) = 0;
}

void Hex8::buildL_thermic(gmm::dense_matrix<double> &L_e, gmm::dense_matrix<double> &K_conductivity)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(L_e); // dimension: 8*8
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);
    std::vector<double> factor3(3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            gmm::mult(K_conductivity, gmm::col_vector(factor1), gmm::col_vector(factor2));
            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor3));
                L_e(i, j) += gmm::vect_sp(factor2, factor3) * gauss.w[a] * detJ;
            }
        }
    }

    for (size_t i = 0; i < nodes.size(); ++i)
        for (size_t j = i; j < nodes.size(); ++j)
        {
            if (fabs(L_e(i, j)) < epsilon)
                L_e(i, j) = 0;
            L_e(j, i) = L_e(i, j);
        }
}

void Hex8::strain_stress(gmm::dense_matrix<double> &Strain, gmm::dense_matrix<double> &Stress, gmm::dense_matrix<double> &H, std::vector<double> &u)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(Strain);
    gmm::clear(Stress);

    gmm::dense_matrix<double> Tmp(3, 3);
    gmm::clear(Tmp);

    gmm::dense_matrix<double> du_d_Phi(3, 3);
    gmm::dense_matrix<double> u_e(3, nodes.size());

    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        gmm::clear(du_d_Phi);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            for (auto k = 0; k < 3; ++k)
                for (auto j = 0; j < 3; ++j)
                    du_d_Phi(j, k) += dffi[j] * u_e(k, i);
        }

        for (size_t k = 0; k < 3; ++k)
            for (size_t m = 0; m < 3; ++m)
                for (size_t i = 0; i < 3; ++i)
                    Tmp(k, m) += J(k, i) * du_d_Phi(i, m) / (nodes.size());
    }
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            Strain(i, j) = (Tmp(i, j) + Tmp(j, i)) / 2;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            if (fabs(Strain(i, j)) < epsilon)
                Strain(i, j) = 0;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            if (fabs(Stress(i, j)) < epsilon)
                Stress(i, j) = 0;
}

void Hex8::buildM(gmm::dense_matrix<double> &M)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    gmm::clear(M);

    gmm::dense_matrix<double> Mk(nodes.size(), nodes.size());

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        //Pt ksi(gauss.p[k]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        gmm::mult(gmm::col_vector(ff), gmm::row_vector(ff), Mk); // Mk = {ff}^T [ff]    // idem pour tous les elements/pg "k" => cache?
        gmm::add(gmm::scaled(Mk, gauss.w[k] * detJ), M);         // M = M + Mk.wG.detJ
    }
}

void Hex8::build(MATTYPE type, gmm::dense_matrix<double> &A, std::vector<double> const &u, Fct0 const &fct)
{
    switch (type)
    {
    case MATTYPE::MAT_K:
        buildK(A, u, fct);
        break;
    case MATTYPE::MAT_M:
        buildM(A);
        break;
    default:
        std::stringstream out;
        out << "Hex8 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute intergal of function
 * I = int fct dV
 * @authors Adrien Crovato
 */
double Hex8::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheHex8 &cache = getCache();
    GaussHex8 &gauss = cache.gauss;
    MemHex8 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Hex8::computeV()
{
    MemHex8 &mem = getMem();
    return mem.getVol();
}
