/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wQuad4.h"
#include "wSfQuad4.h"
#include "wGaussQuad4.h"
#include "wCacheQuad4.h"
#include "wMemQuad4.h"
#include "wNode.h"
#include "wMath.h"
#include "wLazy.h"
#include "wFct0.h"
#include "gmm_gmm.h"

using namespace tbox;

Quad4::Quad4(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemQuad4(*this, static_cast<int>(getCache().gauss.p.size()));
}
Quad4::~Quad4()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Quad4 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Quad4::getVMem()
{
    return getMem();
}
MemQuad4 &Quad4::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Quad4 
 * @authors Romain Boman, Adrien Crovato
 */
CacheQuad4 &Quad4::getCache()
{
    static Lazy<CacheQuad4> cache(2);
    return cache.get();
}
Cache &Quad4::getVCache()
{
    return getCache();
}

Pt Quad4::normal()
{
    Pt x1 = nodes[0]->pos;
    Pt x2 = nodes[1]->pos;
    Pt x3 = nodes[2]->pos;
    Pt x4 = nodes[3]->pos;

    return ((x3 - x1) ^ (x4 - x1)).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 * @authors Romain Boman
 */
double Quad4::buildJ(size_t k) const
{
    CacheQuad4 &cache = getCache();

    std::vector<double> tg1(3), tg2(3);
    gmm::clear(tg1);
    gmm::clear(tg2);

    int i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];

        gmm::add(gmm::scaled((*it)->pos.x, dff[0]), tg1);
        gmm::add(gmm::scaled((*it)->pos.x, dff[1]), tg2);
    }

    std::vector<double> n(3);
    cross(tg1, tg2, n);
    double detJ = gmm::vect_norm2(n);
    return detJ;
}
/**
 * @brief Compute Jacobian matrix at Gauss point k
 * @authors Romain Boman
 */
double Quad4::buildJ(size_t k, gmm::dense_matrix<double> &J) const
{
    //std::cout << "\tTri3::buildJ...\n";
    CacheQuad4 &cache = getCache();

    gmm::dense_matrix<double> Jn(2, 2);
    gmm::clear(J);

    int i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        //std::cout << "\ti=" << i << std::endl;
        std::vector<double> &dff = cache.dff[k][i];
        gmm::mult(gmm::col_vector(dff), gmm::row_vector(gmm::sub_vector((*it)->pos.x, gmm::sub_interval(0, 2))), Jn); // x is 3D so we take the first 2 componenents (x and y)
        gmm::add(Jn, J);
    }
    double detJ = gmm::lu_det(J);
    return detJ;
}

void Quad4::buildS(gmm::dense_matrix<double> &S)
{
    CacheQuad4 &cache = getCache();
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    gmm::clear(S);

    gmm::dense_matrix<double> Sk(nodes.size(), nodes.size());

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        gmm::mult(gmm::col_vector(ff), gmm::row_vector(ff), Sk); // Sk = {ff}^T [ff]    // idem pour tous les elements/pg "k" => cache?
        gmm::add(gmm::scaled(Sk, gauss.w[k] * detJ), S);         // S = S + Sk.wG.detJ
    }
}

void Quad4::build(MATTYPE type, gmm::dense_matrix<double> &A)
{
    switch (type)
    {
    case MATTYPE::MAT_S:
        buildS(A);
        break;
    default:
        std::stringstream out;
        out << "Quad4 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Romain Boman
 */
double Quad4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheQuad4 &cache = getCache();
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    double V = 0.0;

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Quad4::computeV()
{
    MemQuad4 &mem = getMem();
    return mem.getVol();
}