/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSfHex8.h"
#include <stdexcept>

using namespace tbox;

void SfHex8::eval(Pt const &ksi, std::vector<double> &ff) const
{
    ff.resize(8);
    double x = ksi.x[0];
    double y = ksi.x[1];
    double z = ksi.x[2];

    ff[0] = 0.125 * (1.0 + x) * (1.0 + y) * (1.0 - z);
    ff[1] = 0.125 * (1.0 - x) * (1.0 + y) * (1.0 - z);
    ff[2] = 0.125 * (1.0 - x) * (1.0 - y) * (1.0 - z);
    ff[3] = 0.125 * (1.0 + x) * (1.0 - y) * (1.0 - z);
    ff[4] = 0.125 * (1.0 + x) * (1.0 + y) * (1.0 + z);
    ff[5] = 0.125 * (1.0 - x) * (1.0 + y) * (1.0 + z);
    ff[6] = 0.125 * (1.0 - x) * (1.0 - y) * (1.0 + z);
    ff[7] = 0.125 * (1.0 + x) * (1.0 - y) * (1.0 + z);
}

void SfHex8::evalD(int i, Pt const &ksi, std::vector<double> &dff) const
{
    dff.resize(3);
    double x = ksi.x[0];
    double y = ksi.x[1];
    double z = ksi.x[2];

    switch (i)
    {
    case 0:
        dff[0] = 0.125 * (1.0 + y) * (1.0 - z);
        dff[1] = 0.125 * (1.0 + x) * (1.0 - z);
        dff[2] = -0.125 * (1.0 + x) * (1.0 + y);
        break;
    case 1:
        dff[0] = -0.125 * (1.0 + y) * (1.0 - z);
        dff[1] = 0.125 * (1.0 - x) * (1.0 - z);
        dff[2] = -0.125 * (1.0 - x) * (1.0 + y);
        break;
    case 2:
        dff[0] = -0.125 * (1.0 - y) * (1.0 - z);
        dff[1] = -0.125 * (1.0 - x) * (1.0 - z);
        dff[2] = -0.125 * (1.0 - x) * (1.0 - y);
        break;
    case 3:
        dff[0] = 0.125 * (1.0 - y) * (1.0 - z);
        dff[1] = -0.125 * (1.0 + x) * (1.0 - z);
        dff[2] = -0.125 * (1.0 + x) * (1.0 - y);
        break;
    case 4:
        dff[0] = 0.125 * (1.0 + y) * (1.0 + z);
        dff[1] = 0.125 * (1.0 + x) * (1.0 + z);
        dff[2] = 0.125 * (1.0 + x) * (1.0 + y);
        break;
    case 5:
        dff[0] = -0.125 * (1.0 + y) * (1.0 + z);
        dff[1] = 0.125 * (1.0 - x) * (1.0 + z);
        dff[2] = 0.125 * (1.0 - x) * (1.0 + y);
        break;
    case 6:
        dff[0] = -0.125 * (1.0 - y) * (1.0 + z);
        dff[1] = -0.125 * (1.0 - x) * (1.0 + z);
        dff[2] = 0.125 * (1.0 - x) * (1.0 - y);
        break;
    case 7:
        dff[0] = 0.125 * (1.0 - y) * (1.0 + z);
        dff[1] = -0.125 * (1.0 + x) * (1.0 + z);
        dff[2] = 0.125 * (1.0 + x) * (1.0 - y);
        break;
    default:
        throw std::runtime_error("bad index (SfHex8)!");
    }
}
