/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wRungeKutta.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTimer.h"
#include "wTag.h"
#include "wProblem.h"
#include "wDisplayHook.h"
#include "wSource.h"
#include "wResults.h"
#include "wMshExport.h"
#include "gmm_gmm.h"
using namespace tbox;
using namespace waves;

RungeKutta::RungeKutta(std::shared_ptr<waves::Problem> _pbl) : TimeIntegration(_pbl)
{
}

void RungeKutta::start(std::shared_ptr<MshExport> mshWriter)
{
    std::cout << *this;
    chrono1.start();

    auto msh = pbl->msh;
    u1.resize(msh->nodes.size());
    v1.resize(msh->nodes.size());

    // assign a row for each node
    int i = 0;
    for (auto n : msh->nodes)
        n->row = i++;

    // matrix assembly
    gmm::csr_matrix<double> K;
    std::vector<double> Md(msh->nodes.size());
    buildKM(K, Md, u1);

    gmm::csr_matrix<double> S;
    buildS(S);

    // setup results
    Results results;
    results.scalars_at_nodes["u"] = &u1;
    //results.scalars_at_nodes["v"] = &v1;

    // time integration

    std::cout << "\n\nTime integration...\n";
    std::cout << *this << '\n';

    std::vector<double> tmp1(msh->nodes.size()), tmp2(msh->nodes.size());
    double t = 0.0;
    int nt = 1;
    results.nt = nt;
    results.time = t;

    std::vector<double> up1(msh->nodes.size());
    std::vector<double> vp1(msh->nodes.size());
    std::vector<double> up2(msh->nodes.size());
    std::vector<double> vp2(msh->nodes.size());
    std::vector<double> up3(msh->nodes.size());
    std::vector<double> vp3(msh->nodes.size());

    std::vector<double> u2(msh->nodes.size());
    std::vector<double> v2(msh->nodes.size());

    //Node *snod = msh->ntags["Source Point"]->elems[0]->nodes[0];

    while (t < ttot)
    {
        std::cout << "nt " << nt << " t=" << t << "\n";

        //u1[snod->row] = sin(2*M_PI*20*t); // BC
        //if(source) source->apply(t, u1);
        std::for_each(pbl->srcs.begin(), pbl->srcs.end(), [&](std::shared_ptr<Source> s) { s->apply(t, u1); });

        t += dt;
        nt++;

        predictor(up1, vp1, u1, v1, u1, v1, tmp1, tmp2, dt, Md, K, S, 1.0, t);
        predictor(up2, vp2, up1, vp1, u1, v1, tmp1, tmp2, dt, Md, K, S, 1.0, t);
        predictor(up3, vp3, up2, vp2, u1, v1, tmp1, tmp2, dt, Md, K, S, 1.0, t);

        predictor(u2, v2, u1, v1, u1, v1, tmp1, tmp2, dt, Md, K, S, 1. / 2., t);
        predictor(u1, v1, up1, vp1, u2, v2, tmp1, tmp2, dt, Md, K, S, 1. / 3., t);
        predictor(u2, v2, up2, vp2, u1, v1, tmp1, tmp2, dt, Md, K, S, 1. / 8., t);
        predictor(u1, v1, up3, vp3, u2, v2, tmp1, tmp2, dt, Md, K, S, 1. / 24., t);

        //u1[snod->row] = sin(2*M_PI*20*t); //BC bis
        //if(source) source->apply(t, u1);
        std::for_each(pbl->srcs.begin(), pbl->srcs.end(), [&](std::shared_ptr<Source> s) { s->apply(t, u1); });

        if (nt % savefreq == 0)
        {
            //std::cout << "nt " << nt << " t=" << t << "\n";
            results.time = t;
            results.nt++;
            //mshWriter->save("res", results);

            if (dhook)
                dhook->display(nt, t, u1);
        }
        if (dhook)
            dhook->refresh();
    }

    chrono1.stop();
    std::cout << "[cpu] " << chrono1 << '\n';
}

void RungeKutta::predictor(std::vector<double> &up2, std::vector<double> &vp2, // premier membre
                           std::vector<double> &up1, std::vector<double> &vp1, // predicteur
                           std::vector<double> &u0, std::vector<double> &v0,   // sol au temps precedent
                           std::vector<double> &tmp1, std::vector<double> &tmp2, double dt,
                           std::vector<double> &Md,
                           gmm::csr_matrix<double> &K,
                           gmm::csr_matrix<double> &S, double ak, double t)
{
    // equation #1
    gmm::add(u0, gmm::scaled(vp1, ak * dt), up2);

    //up2[snod->row] = sin(2*M_PI*20*t); //BC bis
    //if(source) source->apply(t, up2);

    std::for_each(pbl->srcs.begin(), pbl->srcs.end(), [&](std::shared_ptr<Source> s) { s->apply(t, up2); });

    // equation #2
    gmm::mult(K, up1, tmp1);
    gmm::mult(S, vp1, tmp2);
    for (size_t i = 0; i < gmm::vect_size(Md); ++i)
        vp2[i] = v0[i] - ak * dt / Md[i] * (tmp1[i] + tmp2[i]);
}
